package othello.observation;

import othello.util.Assert;

/** This is a readers and writers lock.
 * There may only be one writer at a time.  The writer is either active or passive.
 * There may only be readers if there is either no writer, or the writer is passive.
 * Only active writers should change the state of the object protected by this object.
 * Usage.
 * <pre>
 * class ProtectedObject {
 *     private ReadersAndWriters rw = new ReadersAndWriters() ;
 *     
 *     public mutator() {
 *         rw.writerEnters() ;
 *         try {
 *             make changes
 *             rw.writerGoesPassive() ;
 *             notify any observers
 *             rw.writerGoesActive() ;
 *             make more changes
 *             rw.writerGoesPassive() ;
 *             notify any observers
 *             rw.writerGoesActive() ; }
 *         finally { rw.writerExits() ; } }
 *     
 *     public accessor() {
 *         rw.readerEnters() ;
 *         try {
 *             access the data }
 *         finally { rw.readerExits() ; }
 * }
 * </pre>
 * @author Theo
 *
 */
class ReadersAndWriters {
	private int numberOfReaders = 0 ;
	private int numberOfWriters = 0 ;
	private boolean writerIsActive = false ;
	
	protected void invariant() {
		Assert.check( numberOfWriters==0 && writerIsActive==false || numberOfWriters==1 ) ;
		Assert.check( numberOfReaders >= 0 ) ;
		Assert.check( numberOfWriters==0 || numberOfReaders==0 || !writerIsActive ) ;
	}
	
	public synchronized void readerEnters() {
		invariant() ;
		while( numberOfWriters > 0 && writerIsActive )
			try { wait() ;
			} catch (InterruptedException e) {
				e.printStackTrace(); }
		invariant() ;
		numberOfReaders++ ;
		invariant() ;
	}
	
	public synchronized void readerExits() {
		invariant() ;
		Assert.check( (numberOfWriters==0 || ! writerIsActive)
				   && numberOfReaders > 0 ) ;
		numberOfReaders-- ;
		notifyAll() ;
		invariant() ;
	}
	
	public synchronized void writerEnters() {
		invariant() ;
		while( numberOfReaders + numberOfWriters > 0 )
			try { wait() ;
			} catch (InterruptedException e) {
				e.printStackTrace(); }
		invariant() ;
		numberOfWriters ++ ;
		writerIsActive = true ;
		invariant() ;
	}
	
	public synchronized void writerExits() {
		invariant() ;
		Assert.check( numberOfWriters == 1 ) ;
		numberOfWriters -- ;
		writerIsActive = false ;
		notifyAll() ;
		invariant() ;
	}
	
	public synchronized void writerGoesPassive() {
		invariant() ;
		Assert.check( numberOfWriters == 1  ) ;
		writerIsActive = false ;
		notifyAll() ;
		invariant() ;
	}
	
	public synchronized void writerGoesActive() {
		invariant() ;
		Assert.check( numberOfWriters == 1  ) ;
		while( numberOfReaders > 0 )
			try { wait() ;
			} catch (InterruptedException e) {
				e.printStackTrace(); }
		invariant() ;
		writerIsActive = true ;
		invariant() ;
	}
}